import http.cookies
import json
import os
import urllib.parse
import traceback
from functools import update_wrapper

def qsp(event):
    if not event:
        return None
    return event.get("queryStringParameters") if event.get("queryStringParameters") else {}

def base_path(event):
    if not event:
        return None
    pathlen = len(event["path"].rstrip("/"))
    if pathlen:
        return event["requestContext"]["path"][:-1*pathlen].rstrip("/")
    else:
        return event["requestContext"]["path"].rstrip("/")

def domain(event, prefix="https://"):
    if not event:
        return None
    return (prefix if prefix else "") + event["headers"]["Host"]

def base_url(event, prefix="https://"):
    if not event:
        return None
    return domain(event, prefix=prefix) + base_path(event)

def static_path(event):
    if "STATIC_BUCKET" in os.environ and "STATIC_PATH" in os.environ:
        return "https://s3.amazonaws.com/{STATIC_BUCKET}/{STATIC_PATH}".format(**os.environ)
    if not event:
        return None
    return base_path(event)



import base64
import json
import os
from functools import update_wrapper
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
from sneks.sam.response_core import make_response, redirect, ResponseException
from sneks.sam import events
from sneks import snekjson
import traceback

env = Environment(loader=FileSystemLoader(os.path.join(os.environ["LAMBDA_TASK_ROOT"], "jinja_templates")))

def add_jinja_filter(name, func):
    env.filters[name] = func

conf_files = ["static_config.json", "extra_params.json"]
for filename in conf_files:
    try:
        with open(os.path.join(os.environ["LAMBDA_TASK_ROOT"], filename)) as f:
            data = json.load(f)
            for k in data:
                os.environ[k] = data[k].strip()
    except:
        traceback.print_exc()
        print("Unable to load configuration file '{}'".format(filename))

content_types = {
    "ico":"image/x-icon",
    "jpg":"image/jpg",
    "jpeg":"image/jpeg",
    "png":"image/png",
    "gif":"image/gif",
    "bmp":"image/bmp",
    "tiff":"image/tiff",
    "txt":"text/plain",
    "rtf":"application/rtf",
    "ttf":"font/ttf",
    "css":"text/css",
    "html":"text/html",
    "js":"application/javascript",
    "eot":"application/vnd.ms-fontobject",
    "svg":"image/svg+xml",
    "woff":"application/x-font-woff",
    "woff2":"application/x-font-woff",
    "otf":"application/x-font-otf",
    "json":"application/json",
    }

binary_types = [
    "image/jpg",
    "image/jpeg",
    "image/png",
    "image/gif",
    "image/bmp",
    "image/tiff",
    "image/x-icon",
    "font/ttf",
    "application/vnd.ms-fontobject",
    "application/x-font-woff",
    "application/x-font-otf",
]

def get_content_type(fname, body=None):
    return content_types.get(fname.split(".")[-1].lower(),"binary/octet-stream")

def get_params(template_name, event=None, **kwargs):
    params = {}
    params["this_page"] = template_name
    if event:
        params["event"] = event
        params["base_path"] = events.base_path(event)
        params["static_base"] = events.static_path(event)
    params.update(kwargs)
    return snekjson.blob(params)

def get_page(template_name, event=None, **kwargs):
    try:
        return make_response(env.get_template(template_name).render(**get_params(template_name, event, **kwargs)))
    except TemplateNotFound as e:
        traceback.print_exc()
        print(e)
        print(dir(e))
        print(vars(e))
        return make_404(event)

def render_page(name, params={}, code=200, headers={}):
    params = params if params else {}
    headers = headers if headers else {}

    template = env.get_template(name)
    body = template.render(**params)
    _headers = {"Content-Type": get_content_type(name)}
    _headers.update(headers)
    return make_response(body=body, code=code, headers=_headers, is_base64=False)

def is_response(d):
    return len(d) == 4 and "body" in d and "statusCode" in d and "headers" in d and "isBase64Encoded" in d

def loader_for(template_name):
    def new_decorator(func):
        def newfunc(event, *args, **kwargs):
            params = func(event, *args, **kwargs)
            if is_response(params):
                return params
            if "this_page" not in params:
                params["this_page"] = template_name
            return get_page(template_name, event, **params)
        update_wrapper(newfunc, func)
        return newfunc
    return new_decorator

def make_message(message, heading="Example Website", event=None, **kwargs):
    template_name = "message.html.jinja"
    body = env.get_template(template_name).render(message_title=heading, message_html=message, **get_params(template_name, event))
    return make_response(body, **kwargs)

def get_static(filename, event=None, **kwargs):
    if "STATIC_BUCKET" in os.environ and "STATIC_PATH" in os.environ:
        return redirect("https://s3.amazonaws.com/{STATIC_BUCKET}/{STATIC_PATH}/{filename}".format(filename=filename, **os.environ))
    filepath = os.path.abspath(os.path.join(os.environ["LAMBDA_TASK_ROOT"],filename))
    if os.path.isfile(filepath) and filepath.startswith(os.path.join(os.environ["LAMBDA_TASK_ROOT"],"static","")):
        content_type = get_content_type(filename)
        b64encoded = False
        if content_type in binary_types:
            with open(filepath,"rb") as f:
                body = base64.b64encode(f.read()).decode("utf-8")
                b64encoded = True
        else:
            with open(filepath,"r") as f:
                body = f.read()
        return make_response(body=body, headers={"Content-Type": content_type}, base64=b64encoded)
    else:
        return make_response("404!!1!", code=404)

def get_debug_blob(event):
    templates = env.list_templates()
    return "<pre>\n{}\n\nAvailable Jinja2 templates:\n\n{}\n</pre>".format(snekjson.dumps(event, indent=2, sort_keys=True, ignore_type_error=True).replace(">","&gt").replace("<","&lt"),"\n".join(templates))

def make_debug(event=None, context=None, headers={}, **kwargs):
    templates = env.list_templates()
    return make_response(body=get_debug_blob(event), headers=headers)

def make_404(event=None, context=None, **kwargs):
    raise ResponseException(make_message("<p>I have no idea what you're talking about.</p><br><br><p>{}</p>".format(get_debug_blob(event)), event=event, heading="HTTP/404 !!1!", code=404))



from sneks.sam.response_core import make_response, ResponseException

class HTTPException(ResponseException):
    CODE = 500
    BODY = None
    HEADERS = {}

    @classmethod
    def throw(cls, body=None, code=None, headers=None):
        body = body if body else cls.BODY
        code = code if code else cls.CODE
        headers = headers if headers else cls.HEADERS
        response = make_response(body=body, code=code, headers=headers)
        raise ResponseException(response)

class HTTP400(HTTPException):
    CODE = 400
    BODY = "No can do, buddy."

class HTTP401(HTTPException):
    CODE = 401
    BODY = "No ticket."

class HTTP403(HTTPException):
    CODE = 403
    BODY = "I can't let you do that, Dave."

class HTTP404(HTTPException):
    CODE = 404
    BODY = "I have absolutely no idea what you're talking about."

class HTTP418(HTTPException):
    CODE = 418
    BODY = "I'm a little teapot, short and stout."

class HTTP451(HTTPException):
    CODE = 451
    BODY = "The man says that you can't handle the truth."

class HTTP500(HTTPException):
    CODE = 500
    BODY = "Oops.  We done messed up."

class HTTP501(HTTPException):
    CODE = 501
    BODY = "A day may come when I implement this functionality, but it is not this day."



import boto3
import os

STACK_NAME = os.environ["STACK_NAME"]

STACK = boto3.client("cloudformation").describe_stacks(StackName=STACK_NAME)["Stacks"][0]
STACK_OUTPUTS = {o["OutputKey"]:o["OutputValue"] for o in STACK.get("Outputs",[])}



import json
from decimal import Decimal
import traceback

class blob(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__predefined_attributes__ = [a for a in dir(self)]
        self.__predefined_attributes__.append("__predefined_attributes__")
        for key in self.keys():
            if not key in self.__predefined_attributes__:
                setattr(self, key, self[key])
                pass
            pass
        pass

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if not key in self.__predefined_attributes__:
            setattr(self, key, self[key])
            pass
        pass

    def __delitem__(self, key):
        super().__delitem__(key)
        if not key in self.__predefined_attributes__:
            delattr(self, key)
            pass
        pass

class SnekJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

class SafeSnekJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return SnekJSONEncoder.default(self, obj)
        except TypeError:
            return "<unserializable object of type '{}' with string repr '{}'>".format(type(obj), repr(obj))

def get_kwargs_and_encoder(kwargs):
    if kwargs.get("ignore_type_error") == True:
        del kwargs["ignore_type_error"]
        return kwargs, SafeSnekJSONEncoder
    return kwargs, SnekJSONEncoder

def dump(obj, *args, **kwargs):
    kwargs, encoder = get_kwargs_and_encoder(kwargs)
    return json.dump(make_json_safe(obj), cls=encoder, *args, **kwargs)

def dumps(obj, *args, **kwargs):
    kwargs, encoder = get_kwargs_and_encoder(kwargs)
    return json.dumps(make_json_safe(obj), cls=encoder, *args, **kwargs)

def dumpf(obj, filename, *args, **kwargs):
    with open(filename, "w") as f:
        return dump(obj, f, *args, **kwargs)

def load(*args, **kwargs):
    return json.load(*args, **kwargs)

def loads(*args, **kwargs):
    if isinstance(args[0], dict):
        return args[0]
    return json.loads(*args, **kwargs)

def loadf(filename, *args, **kwargs):
    with open(filename, "r") as f:
        return json.load(f, *args, **kwargs)

def deepload(s):
    while isinstance(s, str):
        try:
            s = loads(s)
        except:
            break
    if isinstance(s, list):
        s = [deepload(x) for x in s]
    if isinstance(s, dict):
        s = {k:deepload(s[k]) for k in s}
    return s

def make_json_safe(item):
    if isinstance(item, list):
        item = [make_json_safe(e) for e in item]
    if isinstance(item, dict):
        item = {k:make_json_safe(item[k]) for k in item}
    if isinstance(item, Decimal):
        item = float(item)
    return item

def make_ddb_safe(item):
    if isinstance(item, list):
        item = [make_ddb_safe(e) for e in item]
    if isinstance(item, dict):
        item = {k:make_ddb_safe(item[k]) for k in item}
    if isinstance(item, float):
        # Python floats only store 53 bits (~15.95 decimal places) of precision, while Decimals are lossless.
        # In order to not store a bunch of useless noise, strip down to just 15 decimal places.
        item = Decimal(format(item, '.15g').rstrip('0'))
    if item is None:
        item = "null"
    if item == "":
        item = json.dumps("")
    return item



import os

def get_repo_name(directory=None):
    here = directory if directory else os.getcwd()
    subdirs = os.listdir(here)
    while len(here) > 1 and '.git' not in subdirs:
        here = os.path.dirname(here)
        subdirs = os.listdir(here)
    if '.git' not in subdirs:
        return None
    return os.path.basename(here)



import http.cookies
import json
import os
import urllib.parse
import traceback
from functools import update_wrapper

def listify(obj):
    if obj == None:
        return []
    if isinstance(obj, (list,tuple)):
        return obj
    return [obj]

def unlistify(obj):
    if obj == None:
        return None
    if isinstance(obj, (list,tuple)):
        if len(obj) == 0:
            return None
        elif len(obj) == 1:
            return obj[0]
    return obj

def listify_dict(d):
    keys = list(d.keys())
    new_d = dict()
    for k in keys:
        new_d[k] = listify(d[k])
    return new_d

def unlistify_dict(d):
    keys = list(d.keys())
    new_d = dict()
    for k in keys:
        new_d[k] = unlistify(d[k])
    return new_d

def dict_append(d, k, v):
    # d = listify_dict(d)
    d[k] = listify(d.get(k,[]))
    d[k].extend(listify(v))
    return d

def update_lists(d1, d2):
    # d1 = listify_dict(d1)
    # d2 = listify_dict(d2)
    for k in d2:
        d1[k] = listify(d1.get(k,[]))
        d1 = dict_append(d1, k, d2[k])
    return d1

def parse_body(body):
    if body.startswith("{"):
        try:
            return json.loads(body)
        except:
            traceback.print_exc()
    else:
        try:
            return urllib.parse.parse_qs(body)
        except:
            traceback.print_exc()
    return {}

def add_event_params(event, *args, **kwargs):
    params = {}
    event["queryStringParameters"] = event.get("queryStringParameters") if event.get("queryStringParameters") else {}
    params.update(event["queryStringParameters"])
    if event["httpMethod"] == "POST" and event.get("body"):
        body = event["body"]
        update_lists(params, parse_body(body))
    cookie_dict = {}
    try:
        cookies = http.cookies.SimpleCookie()
        cookies.load(event["headers"].get("Cookie",""))
        for k in cookies:
            morsel = cookies[k]
            cookie_dict[morsel.key] = morsel.value
    except:
        traceback.print_exc()
    params = listify_dict(params)
    params = {"kwargs":params}
    params["single_kwargs"] = unlistify_dict(params["kwargs"])
    params["cookies"] = cookie_dict
    params["path"] = {}
    params["path"]["raw"] = event["path"]
    params["path"]["base"] = event["requestContext"]["path"][:-1*len(event["path"])].rstrip("/")
    params["path"]["full"] = "https://" + event["headers"]["Host"] + params["path"]["raw"]
    params["path"]["full_base"] = "https://" + event["headers"]["Host"] + params["path"]["base"]
    if "STATIC_BUCKET" in os.environ and "STATIC_PATH" in os.environ:
        params["path"]["static_base"] = "https://s3.amazonaws.com/{STATIC_BUCKET}/{STATIC_PATH}".format(**os.environ)
    else:
        params["path"]["static_base"] = params["path"]["base"]
    params["http"] = {}
    params["http"]["Referer"] = event.get("headers",{}).get("Referer","")
    params["http"]["Referer"] = params["http"]["Referer"] if params["http"]["Referer"] else params["path"]["full_base"]
    params["http"]["User-Agent"] = event.get("headers",{}).get("User-Agent")
    params["http"]["Method"] = event["httpMethod"]
    params["redirect"] = params["single_kwargs"].get("redirect", params["http"]["Referer"])
    params["redirect"] = params["redirect"] if params["redirect"] else params["path"]["full_base"]
    params["objects"] = {}
    event["params"] = params
    return event

def event_params_decorator(func):
    def newfunc(event, *args, **kwargs):
        event = add_event_params(event)
        return func(event, *args, **kwargs)
    update_wrapper(newfunc, func)
    return newfunc
